#include "lexer.h"

#include "token.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <ctype.h>

/* helper functions */
char
advance (struct lexer *l)
{
	/* advances the cur pointer */
	/* (consumes the character) */
	if (!l)
		return 0;

	if (l->cur + 1 > buffer_length (l->source))
		return 0;

	char *code = buffer_get (l->source);
	return code[l->cur++];
}

char
peek (struct lexer *l)
{
	/* just takes a peek at the current character */
	/* (doesn't consume the character) */
	if (!l)
		return 0;

	if (l->cur + 1 > buffer_length (l->source))
		return 0;

	char *code = buffer_get (l->source);
	return code[l->cur];
}

char
lookahead (unsigned int n, struct lexer *l)
{
	/* looks at the next character in the source */
	/* does not consume the character */
	if (!l)
		return 0;

	if (l->cur + n > buffer_length (l->source))
		return 0;

	char *code = buffer_get (l->source);
	return code[l->cur + n];
}

_Bool
match (char expected, struct lexer *l)
{
	/* check if cur matches an expectation */
	/* (consumes the character if the match is true) */
	if (!l)
		return 0;

	if (l->cur + 1 > buffer_length (l->source))
		return 0;

	char *code = buffer_get (l->source);
	if (code[l->cur] != expected)
		return 0;

	l->cur++;
	return 1;
}

void
add_token (enum token_type type, struct lexer *l)
{
	if (!l)
		return;

	unsigned int size = l->cur - l->start;
	char *lexeme = calloc (size, sizeof (char));
	strncpy (lexeme, buffer_get (l->source) + l->start, size);

	vector_push_back (token_create_heap (type, l->line, lexeme), &l->tokens);
	free (lexeme);
}

_Bool
s_eq (const char *s1, const char *s2)
{
	if (!s1 || !s2)
		return 0;

	size_t len1 = strlen (s1);
	size_t len2 = strlen (s2);
	if (len1 != len2)
		return 0;

	return strncmp (s1, s2, len1) == 0;
}

void
handle_number (struct lexer *l)
{
	if (!l)
		return;

	while (isdigit (peek (l)))
		advance (l);

	if (peek (l) == '.' && isdigit (lookahead (1, l)))
		{
			/* if it's a float, consume the second part of the number */
			do
				{
					advance (l);
				}
			while (isdigit (peek (l)));

			add_token (TOK_FLOAT, l);
			return;
		}

	add_token (TOK_INTEGER, l);
}

void
handle_string (char quote, struct lexer *l)
{
	if (!l)
		return;

	while (peek (l) != quote && l->cur < buffer_length (l->source))
		advance (l);

	if (l->cur >= buffer_length (l->source))
		{
			/* we reached the end of the file without closing the string */
			fprintf (stderr,
							 "lexer.c: handle_string - unterminated string on line %d\n",
							 l->line);
			exit (EXIT_FAILURE);
		}

	advance (l); /* consume the closing quote */

	add_token (TOK_STRING, l);
}

void
handle_identifier (struct lexer *l)
{
	if (!l)
		return;

	while (isalnum (peek (l)) || peek (l) == '_')
		advance (l);

	char *slice = buffer_slice (l->start, l->cur, l->source);
	if (s_eq (slice, "if"))
		add_token (TOK_IF, l);
	else if (s_eq (slice, "then"))
		add_token (TOK_THEN, l);
	else if (s_eq (slice, "else"))
		add_token (TOK_ELSE, l);
	else if (s_eq (slice, "true"))
		add_token (TOK_TRUE, l);
	else if (s_eq (slice, "false"))
		add_token (TOK_FALSE, l);
	else if (s_eq (slice, "and"))
		add_token (TOK_AND, l);
	else if (s_eq (slice, "or"))
		add_token (TOK_OR, l);
	else if (s_eq (slice, "while"))
		add_token (TOK_WHILE, l);
	else if (s_eq (slice, "do"))
		add_token (TOK_DO, l);
	else if (s_eq (slice, "for"))
		add_token (TOK_FOR, l);
	else if (s_eq (slice, "func"))
		add_token (TOK_FUNC, l);
	else if (s_eq (slice, "null"))
		add_token (TOK_NULL, l);
	else if (s_eq (slice, "end"))
		add_token (TOK_END, l);
	else if (s_eq (slice, "print"))
		add_token (TOK_PRINT, l);
	else if (s_eq (slice, "println"))
		add_token (TOK_PRINTLN, l);
	else if (s_eq (slice, "ret"))
		add_token (TOK_RET, l);
	else
		add_token (TOK_IDENTIFIER, l);

	free (slice);
}

/* public functions */
struct lexer
lexer_create ()
{
	struct lexer l = { 0 };

	l.tokens = vector_create ();
	l.line = 1;

	return l;
}

void
lexer_set_source (struct buffer *b, struct lexer *l)
{
	if (!b || !l)
		return;

	l->source = b;
}

void
lexer_lex (struct lexer *l)
{
	while (l->cur < buffer_length (l->source))
		{
			l->start = l->cur;

			char c = advance (l);

			/* whitespaces */
			if (c == '\n')
				l->line++;
			else if (c == ' ' || c == '\t' || c == '\r')
				continue;
			/* single-character tokens */
			else if (c == '(')
				add_token (TOK_LPAREN, l);
			else if (c == ')')
				add_token (TOK_RPAREN, l);
			else if (c == '{')
				add_token (TOK_LCURLY, l);
			else if (c == '}')
				add_token (TOK_RCURLY, l);
			else if (c == '[')
				add_token (TOK_LSQUAR, l);
			else if (c == ']')
				add_token (TOK_RSQUAR, l);
			else if (c == '.')
				add_token (TOK_DOT, l);
			else if (c == ',')
				add_token (TOK_COMMA, l);
			else if (c == '+')
				add_token (TOK_PLUS, l);
			else if (c == '-')
				{
					if (match ('-', l))
						{
							while (peek (l) != '\n'
										 && !(l->cur >= buffer_length (l->source)))
								advance (l);

							continue;
						}

					add_token (TOK_MINUS, l);
				}
			else if (c == '*')
				add_token (TOK_STAR, l);
			else if (c == '^')
				add_token (TOK_CARET, l);
			else if (c == '/')
				add_token (TOK_SLASH, l);
			else if (c == ';')
				add_token (TOK_SEMICOLON, l);
			else if (c == '?')
				add_token (TOK_QUESTION, l);
			else if (c == '%')
				add_token (TOK_MOD, l);
			/* multi-character tokens */
			else if (c == '=' && match ('=', l))
				add_token (TOK_EQ, l);
			else if (c == '~')
				add_token (match ('=', l) ? TOK_EQ : TOK_NOT, l);
			else if (c == '<')
				add_token (match ('=', l) ? TOK_LE : TOK_LT, l);
			else if (c == '>')
				add_token (match ('=', l) ? TOK_GE : TOK_GT, l);
			else if (c == ':')
				add_token (match ('=', l) ? TOK_ASSIGN : TOK_COLON, l);
			else if (isdigit (c))
				handle_number (l);
			else if (c == '\'' || c == '"')
				handle_string (c, l);
			else if (isalpha (c) || c == '_')
				handle_identifier (l);
		}
}

void
lexer_print (struct lexer *l)
{
	puts ("Lexer:");

	for (unsigned int i = 0; i < l->tokens.length; i++)
		{
			struct token *t = l->tokens.elements[i];
			printf ("(TOK_TYPE: %d, \"%s\" - at line %d)\n", t->type,
							buffer_get (&t->lexeme), t->line);
		}
}

void
lexer_free (struct lexer *lexer)
{
	if (!lexer)
		return;

	for (unsigned int i = 0; i < lexer->tokens.length; i++)
		{
			struct token *tok = lexer->tokens.elements[i];
			token_free_heap (tok);
		}

	vector_free (&lexer->tokens);
}
