#include "buffer.h"

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

void
buffer_resize (size_t s_len, struct buffer *b)
{
	if (!b)
		return;

	b->capacity += s_len;
	b->buf = realloc (b->buf, sizeof (char) * b->capacity);
	if (!b->buf)
		fputs ("buffer_resize: couldn't resize memory for buffer.", stderr);
}

struct buffer
buffer_create ()
{
	struct buffer b = { 0 };

	b.capacity = 4;
	b.buf = calloc (b.capacity, sizeof (char));
	if (!b.buf)
		{
			fputs ("buffer_create: cannot allocate memory.", stderr);
		}

	return b;
}

void
buffer_append (const char *s, struct buffer *b)
{
	if (!s || !b)
		return;

	size_t s_len = strlen (s);
	if (((b->len + s_len) + 1) >= b->capacity)
		buffer_resize (s_len + 1, b);

	char *pos = (char *)(b->buf + b->len);
	memcpy (pos, s, s_len);
	b->len += s_len;
	b->buf[b->len] = '\0';
}

void
buffer_printf (struct buffer *b, const char *fmt, ...)
{
	va_list list1, list2;

	va_start (list1, fmt);
	va_copy (list2, list1);

	int size = vsnprintf (NULL, 0, fmt, list1);

	va_end (list1);

	char *buf = malloc (sizeof (char) * size + 1);
	vsnprintf (buf, size + 1, fmt, list2);
	va_end (list2);

	buffer_append (buf, b);
	free (buf);
}

char *
buffer_get (struct buffer *b)
{
	if (!b)
		return NULL;

	return b->buf;
}

_Bool
buffer_read (const char *path, struct buffer *b)
{
	if (!path || !b)
		return 0;

	FILE *fp = fopen (path, "r");
	if (!fp)
		{
			fputs ("buffer_read: cannot open path.", stderr);
			return 0;
		}

	fseek (fp, 0, SEEK_END);
	size_t size = ftell (fp);
	fseek (fp, 0, SEEK_SET);

	char *buf = malloc (sizeof (char) * size);
	if (!buf)
		{
			fclose (fp);
			fputs ("buffer_read: cannot allocate memory for temporary buffer.",
						 stderr);
			return 0;
		}
	fread (buf, size, 1, fp);

	buffer_append (buf, b);

	free (buf);
	fclose (fp);

	return 1;
}

_Bool
buffer_eq (const char *s, struct buffer *b)
{
	if (!s || !b)
		return 0;

	size_t slen = strlen (s);
	if (slen != b->len)
		return 0;

	return strncmp (s, b->buf, slen) == 0;
}

char *
buffer_slice (unsigned int start, unsigned int end, struct buffer *b)
{
	if (!b)
		return NULL;

	if (end > buffer_length (b))
		return NULL;

	char *str = calloc ((end - start) + 1, sizeof (char));
	if (!str)
		return NULL;

	char *bstart = buffer_get (b) + start;
	strncpy (str, bstart, end - start);
	str[end - start] = '\0';

	return str;
}

void
buffer_free (struct buffer *b)
{
	if (!b)
		return;

	free (b->buf);
	memset (b, 0, sizeof (struct buffer));
}

size_t
buffer_length (struct buffer *b)
{
	if (!b)
		return 0;

	return b->len;
}
